import logging
from typing import Any, Callable, Dict, Optional, Union

import pandas as pd
from datasets import ClassLabel, Dataset, Features, Value
from pydantic import BaseModel, root_validator
from sentence_transformers import losses
from sentence_transformers.losses.BatchHardTripletLoss import (
    BatchHardTripletLossDistanceFunction,
)
from sklearn import preprocessing
from typing_extensions import TYPE_CHECKING

__LOGGER__ = logging.getLogger(__name__)

if TYPE_CHECKING:
    from setfit import SetFitModel


class SetFitTrainerArgs(BaseModel):
    """
    SetFitTrainerArgs is a Pydantic model that defines the arguments for the SetFitTrainer.
    __NOTE__: it is a simplified version of the official args from the SetFit library.

    train_dataset: Union[dict, Dataset]
    eval_dataset: Union[dict, Dataset] = None
    model_init: Optional[Callable[[], "SetFitModel"]] = None,
    metric: Union[str, Callable[["Dataset", "Dataset"], Dict[str, float]]] = "accuracy",
    metric_kwargs: Optional[Dict[str, Any]] = None,
    loss_class=losses.CosineSimilarityLoss,
    num_iterations: int = 20
    num_epochs: int = 1
    learning_rate: float = 2e-5
    batch_size: int = 16
    seed: int = 42
    column_mapping: dict = None
    use_amp: bool = False
    warmup_proportion: float = 0.1,
    distance_metric: Callable = BatchHardTripletLossDistanceFunction.cosine_distance,
    margin: float = 0.25,
    samples_per_label: int = 2,
    multi_label: bool = False
    labels: list = None
    """

    train_dataset: Union[dict, "Dataset"]
    eval_dataset: Union[dict, "Dataset"] = None
    model_init: Optional[Callable[[], "SetFitModel"]] = None
    metric: Union[str, Callable[["Dataset", "Dataset"], Dict[str, float]]] = "accuracy"
    metric_kwargs: Optional[Dict[str, Any]] = None
    loss_class = losses.CosineSimilarityLoss
    num_iterations: int = 20
    num_epochs: int = 1
    learning_rate: float = 2e-5
    batch_size: int = 16
    seed: int = 42
    column_mapping: dict = None
    use_amp: bool = False
    warmup_proportion: float = 0.1
    distance_metric: Callable = BatchHardTripletLossDistanceFunction.cosine_distance
    margin: float = 0.25
    samples_per_label: int = 2
    multi_label: bool = False
    labels: list = None

    class Config:
        arbitrary_types_allowed = True
        fields = {"multi_label": {"exclude": True}, "labels": {"exclude": True}}

    @root_validator
    def convert_dict_to_dataset(cls, values):
        def _convert_dict_to_dataset(ds_dict):
            df = pd.DataFrame(datasets_dict)
            labels = df.label.unique().tolist()
            df_duplicate_in_group = df.drop_duplicates(subset=["text", "label"])
            df_duplicates_across_groups = df.drop_duplicates(
                subset=["text", "label", "split"]
            )
            if (len(df_duplicate_in_group) != len(df)) and (
                len(df_duplicate_in_group) != len(df_duplicates_across_groups)
            ):
                __LOGGER__.warning(
                    "There are duplicate texts acrooss the train and eval data."
                )
            elif len(df_duplicate_in_group) != len(df):
                __LOGGER__.warning("There are duplicate texts in the dataset.")
            df = df.drop_duplicates(subset=["text", "label", "split"])
            df_group = df.groupby("text").agg(list).reset_index()
            if len(df_group) != len(df):
                values["multi_label"] = True
                df = df_group
            return df, labels

        def _create_datasets(df: pd.DataFrame, labels):
            class_label = (
                ClassLabel(names=sorted(labels))
                if labels
                # in case we don't have any labels, ClassLabel fails with Dataset.from_dict({"labels": []})
                else Value(dtype="int64", id=None)
            )

            feature_dict = {
                "text": Value("string"),
                "label": [class_label] if values["multi_label"] else class_label,
            }

            ds = Dataset.from_dict(
                df.to_dict(orient="list"), features=Features(feature_dict)
            )
            ds = ds.shuffle(seed=values["seed"])

            return ds

        def _add_data_to_dict(datasets_dict, data, train_or_test):
            for label, texts in data.items():
                for text in texts:
                    datasets_dict["text"].append(text)
                    datasets_dict["label"].append(label)
                    datasets_dict["split"].append(train_or_test)
            return datasets_dict

        if isinstance(values["train_dataset"], dict):
            options = ["train_dataset"]
            datasets_dict = {"text": [], "label": [], "split": []}
            datasets_dict = _add_data_to_dict(
                datasets_dict, values["train_dataset"], "train_dataset"
            )
            if isinstance(values["eval_dataset"], dict):
                options.append("eval_dataset")
                datasets_dict = _add_data_to_dict(
                    datasets_dict, values["eval_dataset"], "eval_dataset"
                )
            elif isinstance(values["eval_dataset"], Dataset):
                raise ValueError(
                    "train_dataset and eval_dataset must be of the same type"
                )

            df, labels = _convert_dict_to_dataset(datasets_dict)
            values["labels"] = labels

            if values["multi_label"]:
                le = preprocessing.MultiLabelBinarizer()
            else:
                le = preprocessing.LabelEncoder()
            df["label"] = le.fit_transform(df["label"]).tolist()
            text = ""
            for train_or_test in options:
                if values["multi_label"]:
                    df_filtered = df.copy(deep=True)
                    df_filtered["split"] = df_filtered["split"].apply(
                        lambda x: True if train_or_test in x else False
                    )
                    df_filtered = df_filtered[df_filtered["split"] == True]  # noqa
                else:
                    df_filtered = df[df["split"] == train_or_test]

                if not df_filtered.empty:
                    df_filtered = df_filtered.drop(columns=["split"])
                    values[train_or_test] = _create_datasets(df_filtered, labels)
                    text += f"\n\t{train_or_test}: {len(values[train_or_test])}"

            __LOGGER__.info(
                f"The datasets have been created: \n\tlabels: {values['labels']}\n\tmulti_label: {values['multi_label']}{text}"
            )

            return values
        else:
            return values



import logging
import types

from setfit import SetFitModel, SetFitTrainer
from spacy import util
from spacy.language import Language
from spacy.tokens import Doc

from spacy_setfit.schemas import SetFitTrainerArgs

__LOGGER__ = logging.getLogger(__name__)


class SpacySetFit:
    def __init__(self, nlp: Language, model: SetFitModel, labels=None):
        self.nlp = nlp
        self.model = model
        self.multi_label = self._check_multi_label(model)
        self.labels = labels
        if self.labels:
            self.id2label = {i: label for i, label in enumerate(self.labels)}
        else:
            self.id2label = None

    @staticmethod
    def _check_multi_label(model: SetFitModel):
        if model.multi_target_strategy:
            return True
        else:
            return False

    @staticmethod
    def _from_pretrained(
        pretrained_model_name_or_path: str, setfit_from_pretrained_args: dict = None
    ):
        if setfit_from_pretrained_args is None:
            setfit_from_pretrained_args = {}
        model = SetFitModel.from_pretrained(
            pretrained_model_name_or_path, **setfit_from_pretrained_args
        )
        return model

    @classmethod
    def from_pretrained(
        cls,
        nlp: Language,
        pretrained_model_name_or_path: str,
        setfit_from_pretrained_args: dict = None,
    ):
        model = cls._from_pretrained(
            pretrained_model_name_or_path, setfit_from_pretrained_args
        )
        return cls(nlp, model)

    @classmethod
    def from_trained(
        cls,
        nlp: Language,
        pretrained_model_name_or_path: str,
        setfit_trainer_args: SetFitTrainerArgs,
        setfit_from_pretrained_args: dict = None,
    ):
        setfit_from_pretrained_args[
            "multi_target_strategy"
        ] = setfit_from_pretrained_args.get("multi_target_strategy")
        if setfit_trainer_args.multi_label:
            setfit_from_pretrained_args["multi_target_strategy"] = "one-vs-rest"

        model = cls._from_pretrained(
            pretrained_model_name_or_path, setfit_from_pretrained_args
        )
        trainer = SetFitTrainer(model=model, **setfit_trainer_args.dict())
        trainer.train()
        if setfit_trainer_args.eval_dataset:
            evaluation = trainer.evaluate()
            __LOGGER__.info(f"evaluation: {evaluation}")

        return cls(nlp, model, labels=setfit_trainer_args.labels)

    def _assign_labels(self, doc, prediction: list):
        if self.id2label:
            doc.cats = {
                self.id2label[idx]: float(score) for idx, score in enumerate(prediction)
            }
        else:
            doc.cats = {idx: float(score) for idx, score in enumerate(prediction)}
        return doc

    def __call__(self, doc: Doc):
        """
        It takes a doc, gets the embeddings from the doc, reshapes the embeddings, gets the prediction from the embeddings,
        and then sets the prediction results for the doc

        :param doc: Doc
        :type doc: Doc
        :return: The doc object with the predicted categories and the predicted categories for each sentence.
        """
        if isinstance(doc, str):
            doc = self.nlp(doc)
        prediction = self.model.predict_proba([doc.text])
        doc = self._assign_labels(doc, prediction[0])

        return doc

    def pipe(self, stream, batch_size=128, include_sent=None):
        """
        predict the class for a spacy Doc stream

        Args:
            stream (Doc): a spacy doc

        Returns:
            Doc: spacy doc with ._.cats key-class proba-value dict
        """
        if isinstance(stream, str):
            stream = [stream]

        if not isinstance(stream, types.GeneratorType):
            stream = self.nlp.pipe(stream, batch_size=batch_size)

        for docs in util.minibatch(stream, size=batch_size):
            pred_results = self.model.predict_proba([doc.text for doc in docs])

            for doc, prediction in zip(docs, pred_results):
                yield self._assign_labels(doc, prediction)



import logging

from rich.logging import RichHandler
from spacy.language import Language

from spacy_setfit.models import SpacySetFit
from spacy_setfit.schemas import SetFitTrainerArgs

__all__ = ["SpacySetFit", "SetFitTrainerArgs"]

logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)


@Language.factory(
    "spacy_setfit",
    default_config={
        "pretrained_model_name_or_path": "all-MiniLM-L6-v2",
        "setfit_from_pretrained_args": None,
        "setfit_trainer_args": None,
    },
)
def create_setfit_model(
    nlp: Language,
    name: str,
    pretrained_model_name_or_path: str = "all-MiniLM-L6-v2",
    setfit_from_pretrained_args: dict = None,
    setfit_trainer_args: dict = None,
):
    if setfit_from_pretrained_args is None:
        setfit_from_pretrained_args = {}
    if setfit_trainer_args is None:
        setfit_trainer_args = {}
    else:
        setfit_trainer_args = SetFitTrainerArgs(**setfit_trainer_args)

    if setfit_trainer_args != {}:
        return SpacySetFit.from_trained(
            nlp=nlp,
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            setfit_from_pretrained_args=setfit_from_pretrained_args,
            setfit_trainer_args=setfit_trainer_args,
        )
    else:
        return SpacySetFit.from_pretrained(
            nlp=nlp,
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            setfit_from_pretrained_args=setfit_from_pretrained_args,
        )
